/*
 * Copyright 2007 Gabe Dijkstra and Robert Hensing
 *
 * This file is part of the Huygens Raytracer.
 * The Huygens Raytracer is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * The Huygens Raytracer is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * The GNU General Public License can be found in a file called LICENSE
 * in the directory where this file resides. It is also available at
 * http://www.gnu.org/licenses/
 *
 */

#include "Scene.h"
#include "Camera.h"
#include "Color.h"
#include "Math.h"
#include "Object.h"
#include "Defaults.h"

#include <stdio.h>
#include <vector>

Camera::Camera(Scene *sc, Vector pos) : 
  scene(sc),
  position(pos), direction_transform(NULL), max_light(INFINITY) {}

Camera::~Camera () {
  if (direction_transform)
    delete direction_transform;
}

void
Camera::setDirectionTransformation(Matrix3 t) {
  if (direction_transform)
    delete direction_transform;
  t.normalize();
  direction_transform = new Matrix3(t);
}

void
Camera::setMaxLight(float l) {
  max_light = l;
}

void Camera::setPosition(Vector a) { position = a; }

Scene *Camera::getScene () { return scene; }

void Camera::setMaxDepth (int a) {
  max_depth = a;
}

PlanarCamera::PlanarCamera(Scene *s, const char *str) {
  scene = s;
  dov = 1.0f;
  n = defaults::n_samples;
  int r = sscanf(str, "%*s %*s (%f, %f, %f) %f %i",
		 &position.x, &position.y, &position.z, &dov, &n);
  if (r < 3)
    throw (void*)NULL;
}

void
PlanarCamera::setImageDimensions (int x, int y) {
  width = x;
  inv_width = 1.0f/width;
  half_width = x * 0.5f;
  half_height = y * 0.5f;
}

Color
PlanarCamera::raytracePixel(int x, int y) {
  Color color;
  /* Calculate primary ray */
  int i, j;
  float frac = 1.0f / n;
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      Ray ray;
      
      ray.origin = position;
      ray.direction = Vector(/* maximale waarden van horizontale
			      * component onafhankelijk van beeldbreedte.
			      */
			     (float(x) + frac*i - (half_width)) * inv_width,
			     /* maximale waarden van verticale component
			      * afhankelijk van hoogte/breedte verhouding.
			      * let op: pixel coordinaten en scene
			      * coordinaten zijn verticaal tegengesteld,
			      * dus hier: -y.
			      */
			     ((half_height) - float(y) + frac*j) * inv_width,
			     dov // ``dov", 1/beeldgrootte
			     );
      ray.direction.normalize();
      
      if (direction_transform)
	ray.direction = direction_transform->transform(ray.direction);
      
      Color sample = scene->raytrace(ray, max_depth, NULL);
      sample.clamp(max_light);
      
      color += sample;
    }
  }
  
  color /= float(n*n);

  return color;
}

CylindricCamera::CylindricCamera(Scene *s, const char *str) {
  scene = s;
  n = defaults::n_samples;
  cyl_offset = 0.0f;
  char angle_unit[16];
  int r = sscanf(str, "%*s %*s (%f, %f, %f) %15s %f %f %f %i",
		 &position.x, &position.y, &position.z, angle_unit,
		 &x_start, &x_end, &cyl_offset, &n);
  if (r >= 7) {
    float angle_factor = getAngleFactor(angle_unit);
    x_start *= angle_factor;
    x_end *= angle_factor;
  }
  else 
    throw (void*)NULL;
}

void
CylindricCamera::setImageDimensions(int w, int h) {
  x_step = (x_end - x_start) / w;

  /* using cylinder height: */
  y_step = -x_step;
  y_start = -y_step * 0.5f * h + cyl_offset;
}

Color
CylindricCamera::raytracePixel(int x, int y) {
  Color color;
  int i, j;
  float frac = 1.0f / n;
  
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      float x_angle = x_start + (x + frac * i) * x_step;
      Ray ray;
      
      ray.origin = position;
      ray.direction = Vector(sinf(x_angle), y_start + (y + frac * j) * y_step, cosf(x_angle));
      ray.direction.normalize();
      
      if (direction_transform)
	ray.direction = direction_transform->transform(ray.direction);
      
      Color sample = scene->raytrace(ray, max_depth, NULL);
      sample.clamp(max_light);
      
      color += sample;
    }
  }
  
  color /= float(n*n);

  return color;
}


SphericCamera::SphericCamera(Scene *s, const char *str) {
  char angle_unit[16];
  int r;
  
  scene = s;
  n = defaults::n_samples;
  r = sscanf(str, "%*s %*s (%f, %f, %f) %15s %f %f %f %f %i",
	     &position.x, &position.y, &position.z, angle_unit,
	     &x_start, &x_end, &y_end, &y_start, &n);
  if (r >= 7) {
    float angle_factor = getAngleFactor(angle_unit);
    x_start *= angle_factor;
    x_end *= angle_factor;
  }
  else 
    throw (void*)NULL;
}

void
SphericCamera::setImageDimensions(int w, int h) {
  x_step = (x_end - x_start) / w;
  y_step = (y_end - y_start) / h;
}

Color
SphericCamera::raytracePixel(int x, int y) {
  Color color;
  int i, j;
  float frac = 1.0f / n;
  
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      float x_angle = x_start + (x + frac * i) * x_step;
      float y_angle = y_start + (y + frac * i) * y_step;
      Ray ray;
      
      ray.origin = position;
      ray.direction = Vector(sinf(x_angle), sinf(y_angle), cosf(x_angle));
      ray.direction.normalize();
      
      if (direction_transform)
	ray.direction = direction_transform->transform(ray.direction);
      
      Color sample = scene->raytrace(ray, max_depth, NULL);
      sample.clamp(max_light);
      
      color += sample;
    }
  }
  
  color /= float(n*n);

  return color;
}
