/*
 * Copyright 2007 Gabe Dijkstra and Robert Hensing
 *
 * This file is part of the Huygens Raytracer.
 * The Huygens Raytracer is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * The Huygens Raytracer is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * The GNU General Public License can be found in a file called LICENSE
 * in the directory where this file resides. It is also available at
 * http://www.gnu.org/licenses/
 *
 */

#ifndef Math_h
#define Math_h

#include <cmath>

float getAngleFactor(const char *unitName);

class Vector {
public:
  float x,y,z;
  inline Vector(void): x(0.0f), y(0.0f), z(0.0f) {}
  
  inline Vector(float a, float b, float c): x(a), y(b), z(c) {}

  inline Vector operator+(Vector a) const {
    Vector ret;
    ret.x = x + a.x;
    ret.y = y + a.y;
    ret.z = z + a.z;
    return ret;
  }
  
  inline Vector operator-(Vector a) const {
    Vector ret;
    ret.x = x - a.x;
    ret.y = y - a.y;
    ret.z = z - a.z;
    return ret;
  };
  
  inline Vector operator*(float a) const {
    Vector ret;
    
    ret.x = x * a;
    ret.y = y * a;
    ret.z = z * a;
    
    return ret;
  }
  
  inline Vector operator/(float a) const {
    Vector ret;
    
    ret.x = x / a;
    ret.y = y / a;
    ret.z = z / a;
    
    return ret;
  } 
  
  inline Vector& operator+=(Vector a) {
    x += a.x;
    y += a.y;
    z += a.z;

    return *this;
  }
  
  inline Vector& operator-=(Vector a) {
    x -= a.x;
    y -= a.y;
    z -= a.z;

    return *this;
  }

  inline Vector operator-() const {
    return Vector(-x, -y, -z);
  }
  
  /* Dit kan sneller:
     PPC:
       #include <ppc_intrinsics.h>
       frsqrte (double precision) 
       frqrtes (single precision)
     x86:
       ???
  */
  
  inline float normalize(void) { 
    float len = sqrtf(x*x + y*y + z*z);
    
    x /= len;
    y /= len;
    z /= len;
    
    return len;
  }
      
  inline float length () {
    return sqrtf(x*x + y*y + z*z);
  }
  
  inline friend float dotproduct(Vector a, Vector b) {
    return (a.x * b.x + a.y * b.y + a.z * b.z);
  }

  inline friend Vector crossproduct(Vector a, Vector b) {
    return Vector(a.y * b.z - a.z * b.y,
		  a.z * b.x - a.x * b.z,
		  a.x * b.y - a.y * b.x);
  }
};

class Matrix3 {
public:
  Vector x;
  Vector y;
  Vector z;

  Matrix3();
  inline Matrix3 (Vector tx, Vector ty, Vector tz) : x(tx), y(ty), z(tz) { }
  inline Vector transform (Vector a) {
    return x * a.x + y * a.y + z * a.z;
  }
  void normalize ();
  inline Matrix3 operator *(Matrix3 a) {
    Matrix3 r;
    r.x = a.x * x.x + a.y * x.y + a.z * x.z;
    r.y = a.x * y.x + a.y * y.y + a.z * y.z;
    r.z = a.x * z.x + a.y * z.y + a.z * z.z;
    return r;
  }
  inline Matrix3 &operator *=(Matrix3 a) {
    return *this = *this * a;
  }
};

class Ray {
public:
  Vector origin;
  Vector direction;
  
  inline Ray (Vector o, Vector d): origin(o), direction(d) {} // normalize d svp
  inline Ray (void): origin(Vector()), direction(Vector()) {}
};

#endif
