/*
 * Copyright 2007 Gabe Dijkstra and Robert Hensing
 *
 * This file is part of the Huygens Raytracer.
 * The Huygens Raytracer is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * The Huygens Raytracer is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * The GNU General Public License can be found in a file called LICENSE
 * in the directory where this file resides. It is also available at
 * http://www.gnu.org/licenses/
 *
 */

#include "PNGOutput.h"
#include "Color.h"
#include "Camera.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <png.h>

bool PNGOutput::spinner = 0;

PNGOutput::PNGOutput(Camera *cam, int w, int h, const char *filenameParam) {
  filename = filenameParam;
  width = w;
  height = h;
  camera = cam;
  spinner = strcmp(getenv("TERM"), "dumb");
  brightness = 1.0f;
  gamma = 1.0f;
}

void
PNGOutput::run(void) {
  fprintf(stderr, "PNG Output to %s%s%s,\n  width: %i,\n  height: %i\n",
	  filename?"‘":"stdout",
	  filename?filename:"",
	  filename?"’":"",
	  width, height);
  
  FILE *output = filename?fopen(filename, "wb"):stdout;
  
  if (!output) {
    fprintf(stderr, "PNGOutput: Cannot open output file ‘%s’, exiting.\n", 
	    strerror(errno));
    exit(1);
  }
  
  png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  
  if (!png_ptr) {
    fprintf(stderr, "PNGOutput: png_create_write_struct failed, exiting.\n");
    exit(1);
  }
  
  int compr_level = 7;
  if (getenv("PNG_COMPRESS"))
    sscanf(getenv("PNG_COMPRESS"), "%i", &compr_level);
  
  png_set_compression_level(png_ptr, compr_level);
  
  png_infop info_ptr = png_create_info_struct(png_ptr);
  
  if (!info_ptr) {
    fprintf(stderr, "PNGOutput: png_create_info_struct failed, exiting.\n");
    exit(1);
  }
  
  if (setjmp(png_jmpbuf(png_ptr))) {
    fprintf(stderr, "PNGOutput: Error during init_io, exiting.\n");
    exit(1);
  }
  
  png_init_io(png_ptr, output);
  
  if (setjmp(png_jmpbuf(png_ptr))) {
    fprintf(stderr, "PNGOutput: Error during writing header, exiting.\n");
    exit(1);
  }
  
  /* misschien is 16 bits per component ook interessant */
  png_set_IHDR(png_ptr, info_ptr, width, height,
	       /*bit_depth*/ 8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
	       PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
  
  png_write_info(png_ptr, info_ptr);
  
  if (setjmp(png_jmpbuf(png_ptr))) {
    fprintf(stderr, "PNGOutput: Error during writing bytes, exiting.\n");
    exit(1);
  }
  
  png_bytep row = (png_bytep) malloc(3*width);
  
  if (!row) {
    fprintf(stderr, "PNGOutput: failure to allocate memory for pixel data, exiting.\n");
    exit(1);
  }
  
  if (spinner) fprintf(stderr, "\n");
  
  for (int y = 0;
       y < height;
       y++) {
    for (int x = 0;
	 x < width;
	 x++) {
      Color color = (camera->raytracePixel(x, y) * brightness) ^ gamma;
      color.clamp();
      //      fprintf(stderr, "%i %i %f %f %f %f %f\n", x, y, color.r, color.g, color.b, brightness, gamma);
      row[x*3] = int(color.r * 255);
      row[x*3+1] = int(color.g * 255);
      row[x*3+2] = int(color.b * 255);
    }
    png_write_row(png_ptr, row);
    
    if (spinner) fprintf(stderr, "%f [%*s%*s]\r", (float)y/(float)height, (60*y)/height, "#", 60-(60*y)/height, "");
  }
  
  if (spinner) fprintf(stderr, "%*s\r", 79, " ");
  
  fprintf(stderr, "Klaar.\n");
  
  free(row);

  if (setjmp(png_jmpbuf(png_ptr))) {
    fprintf(stderr, "PNGOutput: Error during end of write, exiting\n");
    exit(1);
  }
  
  png_write_end(png_ptr, NULL);
  
  fclose(output);
}

void
PNGOutput::set_brightness(float b) {
  brightness = b;
}
void
PNGOutput::set_gamma(float g) {
  gamma = g;
}
