/*
 * Copyright 2007 Gabe Dijkstra and Robert Hensing
 *
 * This file is part of the Huygens Raytracer.
 * The Huygens Raytracer is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * The Huygens Raytracer is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * The GNU General Public License can be found in a file called LICENSE
 * in the directory where this file resides. It is also available at
 * http://www.gnu.org/licenses/
 *
 */

#include "Plane.h"
#include "Scene.h"

Plane::Plane (Vector normal, float d, Color color): normal(normal), d(d), color(color) {
  d /= normal.normalize();
  diffuse_coeff = 1.0;
  specular_coeff = 0.0f;
  shininess = 0.0;
  irreflectivity = 1.0f;
}

Plane::Plane (const char *desc) {
  float a,b,c,R,G,B;
  diffuse_coeff = 1.0;
  specular_coeff = 0.0;
  shininess = 10.0;
  reflectivity = 0.0;
  
  int r = sscanf(desc, "%*s [%f, %f, %f] %f {%f, %f, %f} %f %f %f %f",
		 &a, &b, &c, &d, &R, &G, &B, &diffuse_coeff, &specular_coeff, &shininess, &reflectivity);
  
  if (r == 0) // backward compatibility
    r = sscanf(desc, "%*s (%f, %f, %f) %f {%f, %f, %f} %f %f %f %f",
	       &a, &b, &c, &d, &R, &G, &B, &diffuse_coeff, &specular_coeff, &shininess, &reflectivity);
  
  if (r >= 7 && r <= 11) {
    color = Color(R,G,B);
    normal = Vector(a,b,c);
    d /= normal.normalize();
    irreflectivity = 1 - reflectivity;
  }
  else 
    throw (void*)NULL;
}

float
Plane::intersect (Ray ray) {
  float a = dotproduct(normal, ray.direction);
  
  if (a == 0.0f)
    return INFINITY;
  
  float t = (d - dotproduct(normal, ray.origin))
    /*       ----------------------------------  *//
                            a
    ;
  
  if (t > 0.0f)
    return t;
  else
    return INFINITY;
}

Color
Plane::reflect (Ray ray, float t, int recur_depth) {
  Vector point = ray.origin + ray.direction * t;
  Ray refl_ray;
  
  refl_ray.origin = point;
  refl_ray.direction = normal * (-2 * dotproduct(ray.direction, normal)) + ray.direction;
  
  if (reflectivity == 0.0f)
    return scene->shade(this, color, ray, point, normal, recur_depth);
  else 
    return scene->shade(this, color, ray, point, normal, recur_depth) * irreflectivity
      + color * scene->raytrace(refl_ray, recur_depth, this) * reflectivity;
}
