/*
 * Copyright 2007 Gabe Dijkstra and Robert Hensing
 *
 * This file is part of the Huygens Raytracer.
 * The Huygens Raytracer is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * The Huygens Raytracer is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * The GNU General Public License can be found in a file called LICENSE
 * in the directory where this file resides. It is also available at
 * http://www.gnu.org/licenses/
 *
 */

#include "SDLOutput.h"
#include "Color.h"
#include "Camera.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include <SDL_byteorder.h>

SDLOutput::SDLOutput(Camera *cam, int w, int h) :
  brightness(1.0f), gamma(1.0f)
{
  width = w;
  height = h;
  camera = cam;
  cache = new Color[w*h];
  
  screen = SDL_SetVideoMode(width, height, 24, SDL_SWSURFACE);
  
  if (!screen) {
    fprintf(stderr, "SDLOutput: Could not set video mode: %s, exiting.\n", 
	    SDL_GetError());
    exit(1);
  }
  
  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
}

void
SDLOutput::draw_row(char *row, int row_number) {
  SDL_LockSurface(screen);
  
  if (!row || row_number > screen->h - 1) {
    fprintf(stderr, "SDLOutput: Illegal row number or empty row, exiting.\n");
    exit(1);
  }

  char *line = (char *)screen->pixels + row_number * screen->pitch;
  
  memcpy(line, row, width * 3);
  
  SDL_UnlockSurface(screen);
  SDL_UpdateRect(screen, 0, row_number, 0, 1);
}

void
SDLOutput::hdr_redraw () {
  SDL_LockSurface(screen);
  fprintf(stderr, "HDR factor: %f, γ: %f\n", brightness, gamma);
  for (int y = 0; y < height; y++) {
    for (int x = 0; x < width; x++) {
      char *row = &((char*)screen->pixels)[y*screen->pitch];
      Color color = cache[y*width+x] * brightness;
      color ^= gamma;
      color.clamp();
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
      row[x*3+2] = int(color.b * 255);
      row[x*3+1] = int(color.g * 255);
      row[x*3] = int(color.r * 255);
#else
      row[x*3+2] = int(color.r * 255);
      row[x*3+1] = int(color.g * 255);
      row[x*3] = int(color.b * 255);
#endif
    }
  }
  SDL_UnlockSurface(screen);
  SDL_UpdateRect(screen, 0, 0, 0, 0);
}

extern char **argv;

int
SDLOutput::event_loop(void) {
  SDL_Event event;
  bool do_hdr_redraw = 0;
  
  while (SDL_PollEvent(&event)) {
    switch(event.type) {
    case SDL_QUIT:
      return 1;
      break;
    case SDL_KEYDOWN:
      if (event.key.keysym.sym == SDLK_ESCAPE ||
	  event.key.keysym.sym == 'q')
	return 1;
      if (event.key.keysym.sym == ' ')
	execv(argv[0], argv);
      if (event.key.keysym.sym == SDLK_UP)
	brightness *= 1.1, do_hdr_redraw = 1;
      if (event.key.keysym.sym == SDLK_DOWN)
	brightness /= 1.1, do_hdr_redraw = 1;
      if (event.key.keysym.sym == SDLK_RIGHT)
	gamma *= 1.1, do_hdr_redraw = 1;
      if (event.key.keysym.sym == SDLK_LEFT)
	gamma /= 1.1, do_hdr_redraw = 1;
      break;
    case SDL_MOUSEBUTTONDOWN:
      if (event.button.type == SDL_MOUSEBUTTONDOWN) {
	if (event.button.button == SDL_BUTTON_WHEELUP)
	  brightness *= 1.1, do_hdr_redraw = 1;
	else if (event.button.button == SDL_BUTTON_WHEELDOWN)
	  brightness /= 1.1, do_hdr_redraw = 1;
      }
      break;
    default:
      break;
    }
  }
  if (do_hdr_redraw)
    hdr_redraw();
  
  return 0;
}

void
SDLOutput::draw_by_row (void) {
  char *row = (char *)malloc(3 * width);
  int y;
  
  for (y = 0; y < height; y++) {
    for (int x = 0; x < width; x++) {
      Color color = camera->raytracePixel(x, y);
      cache[y*width+x] = color;
      color *= brightness;
      color ^= gamma;
      color.clamp();
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
      row[x*3+2] = int(color.b * 255);
      row[x*3+1] = int(color.g * 255);
      row[x*3] = int(color.r * 255);
#else
      row[x*3+2] = int(color.r * 255);
      row[x*3+1] = int(color.g * 255);
      row[x*3] = int(color.b * 255);
#endif
    }
    draw_row(row, y);

    if (event_loop())
      exit(1);
  }
  
  free(row);
}

void
SDLOutput::draw_progressively (void) {
  int scale, x, y;
  SDL_Rect rect;
  union {
    struct {
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
      Uint8 a,r,g,b;
#else
      Uint8 b,g,r,a;
#endif
    } b;
    Uint32 val;
  } pixel;

  pixel.b.a = 0xff;
  scale = 32;
  
  rect.h = rect.w = scale;
  for (y = 0; y < height; y += scale) {
    rect.y = y;
    SDL_LockSurface(screen);
    for (x = 0; x < width; x += scale) {
      Color color = camera->raytracePixel(x, y);
      for (int j = y; j < y+scale && j < height; j++) {
	for (int i = x; i < x+scale && i < width; i++) {
	  cache[j*width+i] = color;
	}
      }
      color *= brightness;
      color ^= gamma;
      color.clamp();
      rect.x = x;
      
      pixel.b.r = int(color.r * 255);
      pixel.b.g = int(color.g * 255);
      pixel.b.b = int(color.b * 255);
      
      SDL_FillRect(screen, &rect, pixel.val);
    }
    SDL_UnlockSurface(screen);
    SDL_UpdateRect(screen, 0, rect.y, width, rect.h);

    if (event_loop())
      exit(1);
  }

  for (scale /= 2; scale > 0; scale /= 2) {
    rect.h = rect.w = scale;
    for (y = 0; y < height; y += scale) {
      rect.y = y;
      for (x = 0; x < width; x += scale) {
	if (x % (2 * scale) == 0 && y % (2 * scale) == 0)
	  continue;
	Color color = camera->raytracePixel(x, y);
	for (int j = y; j < y+scale && j < height; j++) {
	  for (int i = x; i < x+scale && i < width; i++) {
	    cache[j*width+i] = color;
	  }
	}
	color *= brightness;
	color ^= gamma;
	color.clamp();
	rect.x = x;
	
	pixel.b.r = int(color.r * 255);
	pixel.b.g = int(color.g * 255);
	pixel.b.b = int(color.b * 255);
	
	SDL_FillRect(screen, &rect, pixel.val);
      }
      if (y % 8 == 0) SDL_UpdateRect(screen, 0, rect.y, width, rect.h * 8);

      if (event_loop())
	exit(1);
    }
  }
}

void
SDLOutput::run(void) {
  if (getenv("PROGRESSIVE"))
    draw_progressively();
  else
    draw_by_row();
      
  while(!event_loop())
    usleep(100000);
}

void
SDLOutput::set_brightness(float b) {
  brightness = b;
}

void
SDLOutput::set_gamma(float g) {
  gamma = g;
}
