/*
 * Copyright 2007 Gabe Dijkstra and Robert Hensing
 *
 * This file is part of the Huygens Raytracer.
 * The Huygens Raytracer is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * The Huygens Raytracer is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * The GNU General Public License can be found in a file called LICENSE
 * in the directory where this file resides. It is also available at
 * http://www.gnu.org/licenses/
 *
 */

#include "Scene.h"
#include "Object.h"

#include <vector>

Light::Light(): position(Vector()), color(Color()), positional(0) {}
Light::Light(Vector position): position(position), color(Color()) {}
Light::Light(Vector position, Color color): position(position), color(color) {}
Light::Light(const char *desc) {
  float x,y,z,
    R = 1.0f, G = 1.0f, B = 1.0f, m = 1.0f;
  
  int r = sscanf(desc, "%*s (%f, %f, %f) {%f, %f, %f} * %f", &x, &y, &z, &R, &G, &B, &m);
  if (r == 3 || r == 6 || r == 7) {
    color = Color(R,G,B) * m;
    position = Vector(x,y,z);
    positional = 1;
  }
  else {
    int r = sscanf(desc, "%*s [%f, %f, %f] {%f, %f, %f} * %f", &x, &y, &z, &R, &G, &B, &m);
    if (r == 3 || r == 6 || r == 7) {
      color = Color(R,G,B) * m;
      position = Vector(x,y,z);
      position.normalize();
      positional = 0;
    }
    else
      throw (void*)NULL;
 }
}
Scene::Scene() {
  background = Color(0.0, 0.0, 0.0);
}

Scene::~Scene() {
  std::vector<Object *>::const_iterator iter;

  for (iter = objects.begin(); iter != objects.end(); iter++)
    delete (*iter); 
  
  std::vector<Light *>::const_iterator light;
  
  for (light = lights.begin(); light != lights.end(); light++)
    delete (*light); 
  
}

void
Scene::add(Object *object) {
  object->setScene(this);
  objects.push_back(object);
}

void
Scene::add(Light *light) {
  lights.push_back(light);
}

Color
Scene::raytrace (Ray ray, int recursion_depth, Object *exclude) {
  recursion_depth--;
  
  if (recursion_depth < 0)
    return background;

  float t;
  float t_closest = INFINITY;
  std::vector<Object *>::const_iterator iter;
  Object *closest_obj = NULL;

  for (iter = objects.begin(); iter != objects.end(); iter++) {
    if (*iter != exclude && (t = (*iter)->intersect(ray)) != INFINITY) {
      if (t < t_closest) {
	t_closest = t;
	closest_obj = (*iter);
      }
    }
  }

  if (closest_obj)
     return closest_obj->reflect(ray, t_closest, recursion_depth);
  else
     return background;
}

void
Scene::setAmbient(Color a) { ambient = a; }
void
Scene::setBackground(Color a) { background = a; }

Color
Scene::shade(Object *object, Color color, Ray ray, Vector point, Vector normal, int depth) {
  Color intensity = color * ambient;
  std::vector<Light *>::const_iterator light;
  
  for (light = lights.begin(); light != lights.end(); light++) {
    Vector light_direction;
    float light_t;
    float attenuation_mult;
    bool positional = (*light)->positional;
    
    if (positional) {
      light_direction = (*light)->position - point;
      light_t = light_direction.normalize();
      attenuation_mult = 1.0f/(light_t * light_t);
    }
    else {
      light_direction = (*light)->position;
      light_t = INFINITY;
      attenuation_mult = 1.0f;
    }
    
    Ray shadow_ray;
    shadow_ray.origin = point;
    shadow_ray.direction = light_direction;
    bool intersected = 0;
    
    /* intersection berekenen */
    for (std::vector<Object*>::iterator iter = objects.begin(); iter != objects.end(); iter++) {
      if (*iter == object)
	continue;
      
      float t = (*iter)->intersect(shadow_ray);
      
      if (t != INFINITY && t > 0.0f && t < light_t) {
	intersected = 1;
	break;
      }
    }
    
    if (intersected)
      continue;
    
    float cos_theta = dotproduct(normal, light_direction);
    
    if (cos_theta > 0.0) {
      Color refl_color = (*light)->color * color;
      
      /* diffuse */
      intensity += refl_color * (attenuation_mult * cos_theta * object->diffuse_coeff);
      
      /* specular */
      Vector refl_dir = normal * (-2 * cos_theta) + light_direction;
      refl_dir.normalize();
      
      float a = dotproduct(refl_dir, ray.direction);
      if (a > 0)
	intensity += refl_color * (attenuation_mult * object->specular_coeff * powf(a, object->shininess));
    }
  }
  
  return intensity;
}
