/*
 * Copyright 2007 Gabe Dijkstra and Robert Hensing
 *
 * This file is part of the Huygens Raytracer.
 * The Huygens Raytracer is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * The Huygens Raytracer is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * The GNU General Public License can be found in a file called LICENSE
 * in the directory where this file resides. It is also available at
 * http://www.gnu.org/licenses/
 *
 */

#include "Sphere.h"
#include "Math.h"
#include "Color.h"
#include "Scene.h"

Sphere::Sphere(const char *desc) {
  float x,y,z,
    R = 1.0f, G = 1.0f, B = 1.0f;
  diffuse_coeff = 1.0,
  specular_coeff = 0.25,
  shininess = 5.0;
  reflectivity = 0.0;
  
  int r = sscanf(desc, "%*s (%f, %f, %f) %f {%f, %f, %f} %f %f %f %f", 
		 &x, &y, &z, &radius, &R, &G, &B, 
		 &diffuse_coeff, &specular_coeff, &shininess, &reflectivity);
  
  if (r == 4 || r == 7 || r == 10 || r == 11) {
    color = Color(R,G,B);
    position = Vector(x,y,z);
  }
  else 
    throw (void*)NULL;
}

Sphere::Sphere(void) {
  position = Vector(0.0, 0.0, 0.0);
  radius = 0.1;
  color = Color(0.7, 0.1, 0.2);
  specular_coeff = 0.0f;
}

Sphere::Sphere(Vector pos, float r) {
  position = pos;
  radius = r;
  color = Color(0.2, 0.1, 0.7);
  specular_coeff = 0.0f;
}

Sphere::Sphere(Vector pos, float r, Color c) {
  position = pos;
  radius = r;
  color = c;
  specular_coeff = 0.0f;
}

float
Sphere::intersect (Ray ray) {
  Vector oc = position - ray.origin;
  float doc = dotproduct(oc, oc);
  
  if (doc < (radius * radius))
    return INFINITY;
  
  float tca = dotproduct(oc, ray.direction);

  if (tca < 0)
    return INFINITY;

  float hc = (radius * radius) - doc + (tca * tca);
  
  if (hc < 0)
    return INFINITY;

  return (tca - sqrt(hc));
}

Color
Sphere::reflect (Ray ray, float t, int recur_depth) {
  Vector point = ray.origin + ray.direction * t;
  Vector normal = (point - position) / radius;
  Ray refl_ray;
  refl_ray.origin = point;
  refl_ray.direction = normal * (-2 * dotproduct(ray.direction, normal)) + ray.direction;
  refl_ray.direction.normalize();
  
  if (specular_coeff != 0.0f)
    return scene->shade(this, color, ray, point, normal, recur_depth) * (1 - reflectivity)
      + color * scene->raytrace(refl_ray, recur_depth, this) * reflectivity;
  else
    return scene->shade(this, color, ray, point, normal, recur_depth);
}
