/*
 * Copyright 2007 Gabe Dijkstra and Robert Hensing
 *
 * This file is part of the Huygens Raytracer.
 * The Huygens Raytracer is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * The Huygens Raytracer is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * The GNU General Public License can be found in a file called LICENSE
 * in the directory where this file resides. It is also available at
 * http://www.gnu.org/licenses/
 *
 */

#ifndef NO_SDL
#include "SDLOutput.h"
#endif
#include "PNGOutput.h"
#include "Camera.h"
#include "Scene.h"
#include "Cube.h"
#include "Sphere.h"
#include "Plane.h"
#include "Triangle.h"
#include "Math.h"
#include "Defaults.h"

#include <stdio.h>
#include <stdlib.h>
#include <cstring>
#include <libgen.h>

/* Defaults */

int defaults::n_samples = DEFAULT_N_SAMPLES;

struct Context {
  Scene *scene;
  Camera *camera;
  float brightness;
  float gamma_exp;
  int width;
  int height;
  int maxdepth;
  Matrix3 transformation;
  
  Context() : scene(NULL), camera(NULL),
	      brightness(1.0f), gamma_exp(1.0f),
	      width(DEFAULT_WIDTH), height(DEFAULT_HEIGHT),
	      maxdepth(DEFAULT_RECURSION_LIMIT) { }
};

/*
 * Lijst van objecten waarvan gebruik kan gemaakt worden door de OBJ macro te
 * definieren en vervolgens ALLOBJECTS `aan te roepen'.
 */
#define ALLOBJECTS OBJ(Sphere) OBJ(Plane) OBJ(Triangle)

void
die_usage (void) {
  fprintf(stderr, "\
Usage: \n\
  huygens-raytracer -s SCENE [ -t OUT_TYPE ] [ -o OUT_FILE ] [ -w WIDTH ]\n\
        [ -h HEIGHT ] [ -n DEFAULT_SAMPLES ] [ -g DEF_GAMMA ] [ -b\n\
        DEF_BRIGHTNESS ] [ -r DEFAULT_RECURSION_LIMIT ]\n\
\n\
  OUT_TYPE:  ('display', 'png')\n\
  SCENE:     filename\n\
");
  exit(1);
}

void load_scene(const char *sceneName, Context &ctx) {
  Scene *scene = ctx.scene;
  FILE *file;
  if (strcmp(sceneName, "-") == 0) {
    file = stdin;
  }
  else {
    file = fopen(sceneName, "r");
  }
  
  if (!file)
    fprintf(stderr, "Kon scene bestand (%s) niet openen!\n", sceneName), exit(1);
  
  int lineno = 0;
  char buf[256]; /* lekker ouderwets. */
  
#define FAIL { fprintf(stderr, "Fout op regel %s:%i: \"%s\"\n", sceneName, lineno, buf); exit(1); }
  char *cwd = NULL;
  
  while (fgets(buf, sizeof(buf), file)) {
    lineno++;
    buf[255] = '\0'; /* voor extra zekerheid */
    
    if (buf[0] == '#')
      continue;
    
    char name[256];
    
    if (sscanf(buf, "%256s", name) == 1) {
      if (strcasecmp(name, "cd") == 0) {
	const char *path = buf;
	if (cwd) free(cwd);
	while(!isspace(*path))
	  path++;
	while(isspace(*path))
	  path++;
	
	cwd = strdup(path);
	
	/* line feed char eraf halen */
	char *nl = strrchr(cwd, '\n');
	if (nl)
	  *nl = 0;
      }
      else if (strcasecmp(name, "include") == 0) {
	const char *filename = buf;
	/* filename := einde van `opcode' */
	while(isspace(*filename))
	  filename++;
	while(*filename && !isspace(*filename))
	  filename++;
	while(isspace(*filename))
	  filename++;
	
	/* line feed char eraf halen */
	char *nl = strrchr(filename, '\n');
	if (nl)
	  *nl = 0;
	
	char filenamebuf[256];
	const char *dir;
	char *sceneNameCopy = NULL;
	
	if (!cwd) {
	  sceneNameCopy = strdup(sceneName);
	  dir = dirname(sceneNameCopy); /* dirname gebruikt een statische buffer ergens */
	}
	else
	  dir = cwd;
	snprintf(filenamebuf, sizeof(filenamebuf), "%s/%s", dir, filename);
	if (sceneNameCopy) free(sceneNameCopy);
	load_scene(filenamebuf, ctx);
      }
      if (strcasecmp(name, "camera") == 0) {
	char name2[32];
	int r = sscanf(buf, "%*s %31s", name2);
	if (r != 1) {
	  FAIL;
	}
	else {
#define CAM(camera_name, CameraClass)					\
	  else if (strcmp(name2, camera_name) == 0) {			\
	    try {							\
	      Camera *o = new CameraClass(ctx.scene, (const char *)buf);	\
	      if (ctx.camera) delete ctx.camera;			\
	      ctx.camera = o;						\
	    }								\
	    catch (void *) {						\
	      FAIL;							\
	    }								\
	  }
	  if (0) { } /* macro starts with _else_ */
	  CAM("planar", PlanarCamera)
	    CAM("spheric", SphericCamera)
	    CAM("spherical", SphericCamera)
	    CAM("cylindric", CylindricCamera)
	    CAM("cylindrical", CylindricCamera)
	    /* lang leve de synoniemen */
	  else FAIL;
	}
      }
      else if (strcasecmp(name, "camera-transformation") == 0) {
	char buf2[16];
	Matrix3 m;
	int ret = sscanf(buf, "%*s [%f, %f, %f] [%f, %f, %f] [%f, %f, %f]",
			 &m.x.x, &m.x.y, &m.x.z,
			 &m.y.x, &m.y.y, &m.y.z,
			 &m.z.x, &m.z.y, &m.z.z);
	if (ret == 9)
	  ctx.transformation = m;
	else if (ret == 3) {
	  /* we have a z vector in m.x */
	  Vector z = m.x;
	  z.normalize();
	  ctx.transformation.z = z;
	  ctx.transformation.x.x = z.z;
	  ctx.transformation.x.y = 0;
	  ctx.transformation.x.z = -z.x;
	  ctx.transformation.x.normalize();
	  
	  ctx.transformation.y = crossproduct(z, ctx.transformation.x);
	}
	else if (sscanf(buf, "%*s %15s", buf2) == 1) {
	  if (strcasecmp(buf2, "identity") == 0)
	    ctx.transformation = Matrix3();
	  else
	    FAIL;
	}
	else
	  FAIL;
      }
      else if (strcasecmp(name, "transform-camera") == 0) {
	Matrix3 m;
	int ret = sscanf(buf, "%*s [%f, %f, %f] [%f, %f, %f] [%f, %f, %f]",
			 &m.x.x, &m.x.y, &m.x.z,
			 &m.y.x, &m.y.y, &m.y.z,
			 &m.z.x, &m.z.y, &m.z.z);
	if (ret == 9)
	  ctx.transformation *= m;
	else
	  FAIL;
      }
      else if (strcasecmp(name, "rotate-camera") == 0) {
	Matrix3 m;
	float rx = 0.0f, ry = 0.0f, rz = 0.0f, angle_unit;
	char angle_unit_name[16] = "";
	
	sscanf(buf, "%*s %15s %f %f %f", angle_unit_name, &rx, &ry, &rz);
	angle_unit = getAngleFactor(angle_unit_name);
	rx *= angle_unit;
	ry *= angle_unit;
	rz *= angle_unit;
	
	m.x = Vector(1.0f, 0.0f, 0.0f);
	m.y = Vector(0.0f, cosf(rx), -sinf(rx));
	m.z = Vector(0.0f, sinf(rx), cosf(rx));
	
	ctx.transformation *= m;
	
	m.x = Vector(cosf(ry), 0.0f, sinf(ry));
	m.y = Vector(0.0f, 1.0f, 0.0f);
	m.z = Vector(-sinf(ry), 0.0f, cosf(ry));
	
	ctx.transformation *= m;
	
	m.x = Vector(cosf(rz), -sinf(rz), 0.0f);
	m.y = Vector(sinf(rz), cosf(rz), 0.0f);
	m.z = Vector(0.0f, 0.0f, 1.0f);
	
	ctx.transformation *= m;
	
	ctx.transformation.normalize();
      }
      else if (strcasecmp(name, "ambient") == 0) {
	float r,g,b,l,m = 1.0f;
	int ret = sscanf(buf, "%*s {%f, %f, %f} * %f", &r, &g, &b, &m);
	
	if (ret == 3 || ret == 4) {
	  scene->setAmbient(Color(r, g, b)*m);
	}
	else if (sscanf(buf, "%*s %f", &l) == 1) {
	  scene->setAmbient(Color(l));
	}
	else
	  FAIL;
      }
      else if (strcasecmp(name, "background") == 0) {
	float r,g,b,l;
	
	if (sscanf(buf, "%*s {%f, %f, %f}", &r, &g, &b) == 3) {
	  scene->setBackground(Color(r, g, b));
	}
	else if (sscanf(buf, "%*s %f", &l) == 1) {
	  scene->setBackground(Color(l));
	}
	else
	  FAIL;
      }
      else if (strcasecmp(name, "size") == 0) {
	if (sscanf(buf, "%*s %ix%i", &ctx.width, &ctx.height) == 2 ||
	    sscanf(buf, "%*s %i %i", &ctx.width, &ctx.height) == 2 ||
	    sscanf(buf, "%*s %i, %i", &ctx.width, &ctx.height) == 2) { }
	else 
	  FAIL;
      }
      else if (strcasecmp(name, "gamma") == 0) {
	if (sscanf(buf, "%*s %f", &ctx.gamma_exp) != 1)
	  FAIL;
      }
      else if (strcasecmp(name, "brightness") == 0) {
	if (sscanf(buf, "%*s %f", &ctx.brightness) != 1)
	  FAIL;
      }
#define GEOM(geom_name, geom_class)				\
      else if (strcasecmp(name, geom_name) == 0) {			\
	try {							\
	  Object *o = new geom_class((const char *)buf);	\
	  scene->add(o);					\
	}							\
	catch (void *) {					\
	  FAIL;							\
	}							\
      }
      
#define OBJ(obj) GEOM(#obj, obj)
      ALLOBJECTS
#undef OBJ
	
      else if (strcasecmp(name, "light") == 0) {
	scene->add(new Light((const char *)buf));
      }
    }
  }
}

char **argv;

int
main (int argc, char *_argv[]) {
  int default_width = DEFAULT_WIDTH,
    default_height = DEFAULT_HEIGHT,
    default_maxdepth = DEFAULT_RECURSION_LIMIT;
  float default_gamma_exp = 1.0f,
    default_brightness = 1.0f;
  argv = _argv;

  const char *sceneName = "-",
#ifdef NO_SDL
    *outType = "png",
#else
    *outType = "display",
#endif
    *outFile = NULL;

  enum {
    ARG, SCENE, WIDTH, HEIGHT, OUTFILE, OUTTYPE, SUPERSAMPLING, RECURSIONLIMIT, GAMMA, BRIGHTNESS
  } next = ARG;
  
  for (int i = 1; i < argc; i++) {
    if (next == ARG && argv[i][0] == '-') {
      switch (argv[i][1]) {
      case 'w':	next = WIDTH; break;
      case 'h':	next = HEIGHT; break;
      case 's': next = SCENE; break;
      case 'o': next = OUTFILE; break;
      case 't': next = OUTTYPE; break;
      case 'n': next = SUPERSAMPLING; break;
      case 'r': next = RECURSIONLIMIT; break;
      case 'g': next = GAMMA; break;
      case 'b': next = BRIGHTNESS; break;
      default: die_usage(); break;
      }
    }
    else {
      switch (next) {
      case ARG:	die_usage(); break;
      case SCENE: sceneName = argv[i]; break;
      case WIDTH: if (sscanf(argv[i], "%i", &default_width) != 1) die_usage(); break;
      case HEIGHT: if (sscanf(argv[i], "%i", &default_height) != 1) die_usage(); break;
      case OUTFILE: outFile = argv[i]; break;
      case OUTTYPE: outType = argv[i]; break;
      case SUPERSAMPLING: if (sscanf(argv[i], "%i", &defaults::n_samples) != 1) die_usage(); break;
      case RECURSIONLIMIT: if (sscanf(argv[i], "%i", &default_maxdepth) != 1) die_usage(); break;
      case GAMMA: if (sscanf(argv[i], "%f", &default_gamma_exp) != 1) die_usage(); break;
      case BRIGHTNESS: if (sscanf(argv[i], "%f", &default_brightness) != 1) die_usage(); break;
      }
      next = ARG;
    }
  }
  
  Context ctx;
  ctx.width = default_width;
  ctx.height = default_height;
  ctx.scene = new Scene();
  ctx.camera = NULL;
  
  load_scene(sceneName, ctx);
  
  if (!ctx.camera) {
    fprintf(stderr, "Fout: Er is geen camera gespecificeerd.\n");
    exit(1);
  }
  ctx.camera->setDirectionTransformation(ctx.transformation);
  ctx.camera->setImageDimensions(ctx.width, ctx.height);
  ctx.camera->setMaxDepth(ctx.maxdepth);
  fprintf(stderr, "Zal scene renderen: ‘%s’ %i %i\n", sceneName, ctx.width, ctx.height);
  
#ifndef NO_SDL
  if (std::strcmp(outType, "display") == 0) {
    SDLOutput output(ctx.camera, ctx.width, ctx.height);
    output.set_gamma(ctx.gamma_exp);
    output.set_brightness(ctx.brightness);
    output.run();
  }
  else
#endif
  if (std::strcmp(outType, "png") == 0) {
    ctx.camera->setMaxLight(1.0f/ctx.brightness);
    PNGOutput output(ctx.camera, ctx.width, ctx.height, outFile);
    output.set_gamma(ctx.gamma_exp);
    output.set_brightness(ctx.brightness);
    output.run();
  }
  else {
    fprintf(stderr, "Fout: Specificeer een geldig output type!\n");
    return 1;
  }
  
  return 0;
}
